use std::{alloc::Layout, ffi::{c_char, CStr}, path::PathBuf};

use dlopen::raw::Library;

use crate::{c_str, config::Config};

use self::context::{PluginContext, PluginContextContainer};

pub mod loader;
pub mod context;

#[repr(C)]
pub struct PluginInfo {
    name: c_str,
    version: c_str,
    license: c_str,
}

#[derive(Debug, Default)]
pub struct Plugin {
    syms: Option<loader::PluginSyms>,
    path: PathBuf,
    lib: Option<Library>,
    enabled: bool,
}



impl Plugin {
    pub fn load(path: PathBuf) -> Result<Self, dlopen::Error> {
        let mut s = Self::default();
        s.enabled = true;
        s.path = path;
        s.reload_symbols()?;
        Ok(s)
    }
  
    pub fn syms(&self) -> loader::PluginSyms {
        self.syms.unwrap()
    }

    pub fn init(&self, cfg: &Config) {
        unsafe {
            let conf_dir = &cfg.config_dir.join(self.name()).to_string_lossy().to_string();

            let ctx = PluginContextContainer::new(conf_dir);
            (self.syms().init)(&ctx.inner as *const PluginContext)
        }
    }
    pub fn reload(&mut self) -> Result<(), dlopen::Error> { 
        unsafe {
            let state = (self.syms().pre_reload)();
            if let Err(e) = self.reload_symbols() {
                self.disable();
                return Err(e);
            }
            (self.syms().post_reload)(state);
        }
        Ok(())
    }

    pub fn poll(&self, cap: usize) -> Result<String, std::string::FromUtf8Error> {
        let layout = Layout::from_size_align(cap, 1).unwrap();
        let buf = unsafe {
            std::alloc::alloc_zeroed(layout)
        };
        let s = unsafe {
            (self.syms().poll)(buf as *mut i8, cap);
            let len = libc::strlen(buf as *const i8);
            if len > cap {
                panic!("String len is bigger than allocatd");
            }

            String::from_raw_parts(buf, len, cap)
        };

        if !s.is_empty() {
            log::debug!("Polled: {}", s);
        }
        Ok(s)
    }

    pub fn free(self) {
        unsafe {
            (self.syms().free)();
        }
        drop(self.lib.unwrap()); 
    }

    pub fn enabled(&self) -> bool {
        self.enabled
    }

    pub fn disable(&mut self) {
        self.enabled = false;
    }

    pub fn name(&self) -> String {
        unsafe {
            let info = (self.syms().get_info)();
            CStr::from_ptr((*info).name).to_string_lossy().to_string()
        }
    }

    pub fn version(&self) -> Option<String> {
        unsafe {
            let info = (self.syms().get_info)();
            let v = (*info).version;
            if v.is_null() {
                return None;
            }
            Some(CStr::from_ptr(v).to_string_lossy().to_string())
        }
    }

    pub fn license(&self) -> Option<String> {
        unsafe {
            let info = (self.syms().get_info)();
            let v = (*info).license;
            if v.is_null() {
                return None;
            }
            Some(CStr::from_ptr(v).to_string_lossy().to_string())
        }
    }
}

